classdef DS < handle
    % Paper: "From Pictorial Structures to Deformable Structures", S. Zuffi, O.
    % Freifeld, M. Black, CVPR 2012
    % Authors: Silvia Zuffi, Department of Computer Science, Brown University
    % Contact: zuffi@cs.brown.edu
    % $Date: 2012 $
    % $Revision: $
    %
    % Copyright 2011-2012, Brown University, Providence, RI. USA
    %
    %                          All Rights Reserved
    %
    % All commercial use of this software, whether direct or indirect, is
    % strictly prohibited including, without limitation, incorporation into in
    % a commercial product, use in a commercial service, or production of other
    % artifacts for commercial purposes.
    %
    % Permission to use, copy, modify, and distribute this software and its
    % documentation for research purposes is hereby granted without fee,
    % provided that the above copyright notice appears in all copies and that
    % both that copyright notice and this permission notice appear in
    % supporting documentation, and that the name of the author and Brown
    % University not be used in advertising or publicity pertaining to
    % distribution of the software without specific, written prior permission.
    %
    % For commercial uses contact the Technology Venture Office of Brown University
    %
    % THE AUTHOR AND BROWN UNIVERSITY DISCLAIM ALL WARRANTIES WITH REGARD TO
    % THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
    % FITNESS FOR ANY PARTICULAR PURPOSE.  IN NO EVENT SHALL THE AUTHOR OR
    % BROWN UNIVERSITY BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
    % DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
    % PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
    % ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
    % THIS SOFTWARE.
    
    properties
        nParts = 10;
        nJointsPart = [ 6, 2, 2, 2, 2, 2, 2, 2, 2, 2];
        occlusion = [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
        partColor = ['w', 'y', 'b', 'r', 'b', 'r', 'c', 'm', 'c', 'm'];
        jointColor = 'b';
        linetype = '--';
        lineWidth = 1;
        
        aspect_ratio = 1;
        
        % Current contour pose
        poseAbs;
        
        % Current joints
        Qx, Qy;
        
        % Current part centers
        Ox, Oy;
        
        % Current shape
        z;
        
        % p1 and p2 points
        p1x, p1y, p2x, p2y;
        
        % Contour points
        partX;
        partY;
        
        % Model
        model;
        
        % Bases and mean for each part
        B, M;
        
        % Image
        image;
        imSize;
        scale = 1.2;
        
        
        % indexes of the nodes
        torso = 1;
        head = 2;
        upperRightArm = 3;
        upperLeftArm = 4;
        upperRightLeg = 5;
        upperLeftLeg = 6;
        lowerRightArm = 7;
        lowerLeftArm = 8;
        lowerRightLeg = 9;
        lowerLeftLeg = 10;
        
        % indexes of the models
        headToTorso=1;
        rightArmToTorso=2;
        leftArmToTorso=3;
        rightLegToTorso=4;
        leftLegToTorso=5;
        torsoToHead=6;
        torsoToRightArm=7;
        lowerToUpperRightArm=8;
        torsoToLeftArm=9;
        lowerToUpperLeftArm=10;
        torsoToRightLeg=11;
        lowerToUpperRightLeg=12;
        torsoToLeftLeg=13;
        lowerToUpperLeftLeg=14;
        upperToLowerRightArm=15;
        upperToLowerLeftArm=16;
        upperToLowerRightLeg=17;
        upperToLowerLeftLeg=18;
        nModels = 18;
        
        modelRefDepInd = { [2, 1], [3, 1], [4, 1], [5, 1],...
            [6, 1], [1, 2], [1, 3], [7, 3],...
            [1, 4], [8, 4], [1, 5], [9, 5],...
            [1, 6], [10, 6], [3, 7],...
            [4, 8], [5, 9], [6, 10] };
        
        
        modelName = { 'headToTorso', 'rightArmToTorso', 'leftArmToTorso', 'rightLegToTorso',...
            'leftLegToTorso', 'torsoToHead', 'torsoToRightArm', 'lowerToUpperRightArm',...
            'torsoToLeftArm', 'lowerToUpperLeftArm', 'torsoToRightLeg', 'lowerToUpperRightLeg',...
            'torsoToLeftLeg', 'lowerToUpperLeftLeg', 'upperToLowerRightArm',...
            'upperToLowerLeftArm', 'upperToLowerRightLeg', 'upperToLowerLeftLeg' };
        
        
        nBasis;
        jointConnectionMatrix;
        modelIndex;
        torsoJointsList;
        modelForPartRef;
        flipModelInd;
        Child;
        Parent;
    end
    
    methods
        
        function c = DS( modelName )
            m = load( modelName );
            c.model = m.model;
            
            c.jointConnectionMatrix = [ 0 1 1 1 1 1 0 0 0 0;
                1 0 0 0 0 0 0 0 0 0;
                3 0 0 0 0 0 1 0 0 0;
                4 0 0 0 0 0 0 1 0 0;
                5 0 0 0 0 0 0 0 1 0;
                6 0 0 0 0 0 0 0 0 1;
                0 0 2 0 0 0 0 0 0 0;
                0 0 0 2 0 0 0 0 0 0;
                0 0 0 0 2 0 0 0 0 0;
                0 0 0 0 0 2 0 0 0 0];
            
            c.modelIndex = cell(c.nParts, c.nParts);
            c.modelIndex{c.torso}{c.head} = 'torsoToHead';
            c.modelIndex{c.head}{c.torso} = 'headToTorso';
            c.modelIndex{c.torso}{c.upperRightArm} = 'torsoToRightArm';
            c.modelIndex{c.upperRightArm}{c.torso} = 'rightArmToTorso';
            c.modelIndex{c.torso}{c.upperLeftArm} = 'torsoToLeftArm';
            c.modelIndex{c.upperLeftArm}{c.torso} = 'leftArmToTorso';
            c.modelIndex{c.torso}{c.upperRightLeg} = 'torsoToRightLeg';
            c.modelIndex{c.upperRightLeg}{c.torso} = 'rightLegToTorso';
            c.modelIndex{c.torso}{c.upperLeftLeg} = 'torsoToLeftLeg';
            c.modelIndex{c.upperLeftLeg}{c.torso} = 'leftLegToTorso';
            c.modelIndex{c.upperRightArm}{c.lowerRightArm} = 'upperToLowerRightArm';
            c.modelIndex{c.lowerRightArm}{c.upperRightArm} = 'lowerToUpperRightArm';
            c.modelIndex{c.upperLeftArm}{c.lowerLeftArm} = 'upperToLowerLeftArm';
            c.modelIndex{c.lowerLeftArm}{c.upperLeftArm} = 'lowerToUpperLeftArm';
            c.modelIndex{c.upperRightLeg}{c.lowerRightLeg} = 'upperToLowerRightLeg';
            c.modelIndex{c.lowerRightLeg}{c.upperRightLeg} = 'lowerToUpperRightLeg';
            c.modelIndex{c.upperLeftLeg}{c.lowerLeftLeg} = 'upperToLowerLeftLeg';
            c.modelIndex{c.lowerLeftLeg}{c.upperLeftLeg} = 'lowerToUpperLeftLeg';
            
            c.torsoJointsList = [ c.upperRightArm, c.upperLeftArm, c.upperRightLeg, c.upperLeftLeg ];
            
            c.modelForPartRef = [ c.torsoToHead, c.headToTorso, c.rightArmToTorso, c.leftArmToTorso, c.rightLegToTorso, ...
                c.leftLegToTorso, c.lowerToUpperRightArm, c.lowerToUpperLeftArm, c.lowerToUpperRightLeg, c.lowerToUpperLeftLeg ];
            
            c.flipModelInd = [ c.torsoToHead, c.torsoToRightArm, c.torsoToLeftArm, c.torsoToRightLeg, c.torsoToLeftLeg, c.headToTorso,...
                c.rightArmToTorso, c.upperToLowerRightArm, c.leftArmToTorso, c.upperToLowerLeftArm, c.rightLegToTorso, ...
                c.upperToLowerRightLeg, c.leftLegToTorso, c.upperToLowerLeftLeg, c.lowerToUpperRightArm, c.lowerToUpperLeftArm, ...
                c.lowerToUpperRightLeg, c.lowerToUpperLeftLeg];
            
            c.Child = cell(c.nParts, 1);
            c.Parent = cell(c.nParts, 1);
            
            c.Child{c.torso} = [ c.head, c.upperRightArm, c.upperLeftArm, c.upperRightLeg, c.upperLeftLeg];
            c.Child{c.upperRightArm} = c.lowerRightArm;
            c.Child{c.upperLeftArm} = c.lowerLeftArm;
            c.Child{c.upperRightLeg} = c.lowerRightLeg;
            c.Child{c.upperLeftLeg} = c.lowerLeftLeg;
            
            c.Parent{c.head} = c.torso;
            c.Parent{c.upperRightArm} = c.torso;
            c.Parent{c.upperLeftArm} = c.torso;
            c.Parent{c.upperRightLeg} = c.torso;
            c.Parent{c.upperLeftLeg} = c.torso;
            c.Parent{c.lowerRightArm} = c.upperRightArm;
            c.Parent{c.lowerLeftArm} = c.upperLeftArm;
            c.Parent{c.lowerRightLeg} = c.upperRightLeg;
            c.Parent{c.lowerLeftLeg} = c.upperLeftLeg;
            
            
            c.set_PCA_model();
            try
                c.nBasis = m.nBasis;
            catch
                c.nBasis(1:c.nParts) = size(c.B{1},2);
            end
            [ c.poseAbs c.z  c.Ox c.Oy c.Qx c.Qy c.p1x c.p1y c.p2x c.p2y ] = c.get_mean_pose_shape();
            c.partX = cell(c.nParts,1);
            c.partY = cell(c.nParts, 1);
        end
        
        function set_image( this, image )
            this.image = image;
            this.imSize = deal(size(image));
            this.aspect_ratio = this.imSize(2)/ this.imSize(1);
        end
        
        % Probabilistic model
        [ logP ] = compute_model_log_probability( this, direction );
        [ z Pz ] = sample_part_shape2( this, i, nB );
        [ z Pz ] = sample_part_shape_using_random_stick( this, i, nB, factor );
        [ z_j, theta_j0, Ojx_0, Ojy_0, Qjx_0, Qjy_0, Qx, Qy ] = get_random_child( this, child, node, z_i, theta_i0, Oix_0, Oiy_0, stick, theta_j0);
        [ z_j, sin_jOut, cos_jOut, Ojx_i, Ojy_i, Qjx_i, Qjy_i, mu, C ] = ...
            predict_part_conditioned( this, modelInd, z_i, sin_j, cos_j, stick_j, RANDOM );
        [ mu_ab C_ab zInd sinInd cosInd qxInd qyInd H ] = ...
            get_conditioned_potential( this, refInd, depInd, z_i, sin_ji, cos_ji, stick_j, nBdep );
        [ z_j, sin_j, cos_j, Ojx_i, Ojy_i, Qjx_i, Qjy_i ] = get_part_sample(this, model, Mean, Covariance, z_i);
        
        % Drawing
        draw_full_model( this, fH );
        [ p1x, p1y, p2x, p2y, X, Y ] = draw_part_to_image( this, ind, fH );
        [ p1x p1y p2x p2y X1 Y1 ] = draw_part_to_image_with_parameters( this, ind, z, theta, Ox, Oy, scale, fH, color );
        
        % Geometry
        [ Xi, Yi ] = world_to_image( this, X, Y );
        [ X, Y ] = image_to_world( this, Xi, Yi );
        [ Xi, Yi ] = world_to_image_vector( this, X, Y );
        [ X, Y ] = image_to_world_vector( this, Xi, Yi );
        [ Ox Oy ] = joint_to_part_center( this, ind, z, Angle, Q1x, Q1y, Q2x, Q2y );
        [ Ox Oy ox oy ] = joint_to_part_center_from_index( this, ind, z, Angle, Qx, Qy, index );
        [ Ox Oy pX pY X1 Y1] = joint_to_part_center_joints_points( this, ind, z, Angle, Q1x, Q1y, Q2x, Q2y );
        [ pX pY ] = part_center_to_joints( this, ind, z, theta, Ox, Oy );
        [ Ox Oy pX pY X1 Y1] = joint_to_part_center_joints_points_multiple( this, ind, z, Angle, Q1x, Q1y, Q2x, Q2y );
        [ Xi, Yi ] = world_to_image_with_scale( this, X, Y, scale );
        
        
        % Setting the puppet
        [ P ] = set_me_as_mean( this, Ox, Oy );
        set_DS_as_more_random_puppet( this, startNode, Ox, Oy );
        [ z ] = get_PCAcoefficients_sample(this, partInd, n);
        [ z Pz ] = sample_part_shape( this, i );
        init_with_stickmen( this, stickmen );
        set_FCP_as_random_puppet_from_torso_angle( this, startNode, Ox, Oy, startAngle );
        set_DS_as_random_puppet( this, startNode, modelInd, Ox, Oy )
        set_DS_as_random_puppet_from_pose( this, startNode, Ox, Oy, pose, stick );
        
        
        % Getting info
        [ length ] = get_torso_length( this );
        [ Z ] = get_torso_shape_from_joints( this, Px, Py );
        [ stick stickCoo ] = get_sticks( this );
        [ BBCoo BBCooW dx dy Ox Oy ] = get_bounding_boxes( this, borderSize );
        [ pX pY X1 Y1 ] = get_part_points_local( this, ind, z );
        [ pX pY X1 Y1 ] = get_part_points( this, ind );
        [ pX pY X1 Y1 ] = get_part_points_with_parameters( this, ind, z, theta, Ox, Oy );
        [ poseAbs shape  Ox Oy Qx Qy p1x p1y p2x p2y ] = get_mean_pose_shape( this );
        [ BBCoo, BBCooW, dx, dy, Ox, Oy ] = get_bounding_boxes_from_shape( this, borderSize, shape );
        [ Z C ] = get_torso_from_joints2( this, Px, Py, nB, Angle, Ox, Oy, lambda );
        [ points angles ] = get_part_contour_points_and_angles( this, ind, z, theta, Ox, Oy );
        [ stick stickCoo ] = get_PS_sticks( this );
        [ z_j, theta_j0, Ojx_0, Ojy_0, Qjx_0, Qjy_0, Qx, Qy ] = get_child( this, child, node, z_i, theta_i0, Oix_0, Oiy_0, stick_j, theta_j0)
        
        
        % Generating masks
        [ B ] = make_mask_image( this );
        [ B ] = make_part_mask_image( this, i );
        [ imageBB maskBB angle x1 y1 x2 y2] = get_local_image_BB(this, image, partInd);
        [ O ] = make_mask_image_with_occlusion( this );
        [ B ] = make_part_mask_image_with_parameters( this, i, z, theta, Ox, Oy );
        [ B ] = make_hand_mask_image_with_parameters( this, i, z, theta, Ox, Oy );
        [ B ] = make_part_mask_image_with_parameters_faster( this, i, z, theta, Ox, Oy );
        [ B ] = make_occlusion_mask_with_parameters( this, i, z, theta, Ox, Oy, partPointsScore, jointIndex );
        [ B ] = make_upper_torso_mask_with_parameters( this, z, theta, Ox, Oy );
        [ B ] = make_upper_body_mask_with_parameters( this, z, theta, Ox, Oy, zj, thetaj, Oxj, Oyj );
        
    end
    
end

