
function [ z_j, sin_jOut, cos_jOut, Ojx_i, Ojy_i, Qjx_i, Qjy_i, mu_ab, C_ab ] = ...
    predict_part_conditioned( this, modelInd, z_i, sin_j, cos_j, stick_j, RANDOM )

% PREDICT_PART_CONDITIONED
% Computes parameters for the part j given parameters for the part i, using
% the pairwise model that has i as reference part and j as dependent part.
% If RANDOM is true, returns a random part, otherwise the mean.

% Paper: "From Pictorial Structures to Deformable Structures", S. Zuffi, O.
% Freifeld, M. Black, CVPR 2012
% Authors: Silvia Zuffi, Department of Computer Science, Brown University
% Contact: zuffi@cs.brown.edu
% $Date: 2012 $
% $Revision: $
%
% Copyright 2011-2012, Brown University, Providence, RI. USA
%
%                          All Rights Reserved
%
% All commercial use of this software, whether direct or indirect, is
% strictly prohibited including, without limitation, incorporation into in
% a commercial product, use in a commercial service, or production of other
% artifacts for commercial purposes.
%
% Permission to use, copy, modify, and distribute this software and its
% documentation for research purposes is hereby granted without fee,
% provided that the above copyright notice appears in all copies and that
% both that copyright notice and this permission notice appear in
% supporting documentation, and that the name of the author and Brown
% University not be used in advertising or publicity pertaining to
% distribution of the software without specific, written prior permission.
%
% For commercial uses contact the Technology Venture Office of Brown University
%
% THE AUTHOR AND BROWN UNIVERSITY DISCLAIM ALL WARRANTIES WITH REGARD TO
% THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
% FITNESS FOR ANY PARTICULAR PURPOSE.  IN NO EVENT SHALL THE AUTHOR OR
% BROWN UNIVERSITY BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
% DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
% PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
% ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
% THIS SOFTWARE.


model = this.model;

C = model{modelInd}.Covariance;
mu = model{modelInd}.Mean;
refInd = this.modelRefDepInd{modelInd}(1);
depInd = this.modelRefDepInd{modelInd}(2);
if this.nBasis(refInd) > 0
    z = zeros(this.nBasis(refInd), 1);
    z(1:numel(z_i)) = z_i;
    z_i = z;
else
    z_i = [];
end
nBref = numel(z_i);
nBdep = this.nBasis(depInd);

if this.nBasis(depInd) > 0
    ind = [model{modelInd}.depCoeffInd(1:nBdep), model{modelInd}.sinDepInd, model{modelInd}.cosDepInd, ...
        model{modelInd}.QxDepInd, model{modelInd}.QyDepInd, model{modelInd}.boneLengthDepInd, ...
        model{modelInd}.refCoeffInd(1:nBref) ];
    
    C = C(ind, ind);
    mu = mu(ind);
    
    depCoeffInd = (1:nBdep);
    sinDepInd = depCoeffInd(end)+1;
    cosDepInd = sinDepInd+1;
    QxDepInd = cosDepInd+1;
    QyDepInd = QxDepInd+1;
    boneLengthDepInd = QyDepInd+1;
    refCoeffInd = (boneLengthDepInd+1:boneLengthDepInd+nBref);
else
    ind = [model{modelInd}.sinDepInd, model{modelInd}.cosDepInd, ...
        model{modelInd}.QxDepInd, model{modelInd}.QyDepInd, ...
        model{modelInd}.boneLengthDepInd, model{modelInd}.refCoeffInd(1:nBref) ];
    
    C = C(ind, ind);
    mu = mu(ind);
    depCoeffInd = [];
    sinDepInd = 1;
    cosDepInd = sinDepInd+1;
    QxDepInd = cosDepInd+1;
    QyDepInd = QxDepInd+1;
    boneLengthDepInd = QyDepInd+1;
    refCoeffInd = (boneLengthDepInd+1:boneLengthDepInd+nBref);
end

allInd = (1:size(C,1));
bInd = [];
if numel(z_i)
    bInd = [ bInd, refCoeffInd ];
end
if numel(sin_j) && numel(cos_j)
    bInd = [ bInd, sinDepInd, cosDepInd ];
end
if numel(stick_j)
    bInd = [ bInd, boneLengthDepInd ];
end

aInd = setdiff(allInd, bInd);

conditioningValues = [z_i' sin_j cos_j stick_j];

%% Using the covariance matrix
Caa = C(aInd, aInd);
Cbb = C(bInd, bInd);
Cab = C(aInd, bInd);
Cba = C(bInd, aInd);
if isempty(conditioningValues)
    C_ab = Caa;
    mu_ab = mu(aInd);
else
    C_ab = Caa - Cab * (Cbb \ Cba);
    mu_ab = mu(aInd) + Cab * (Cbb \ (conditioningValues' - mu(bInd)));
end

if RANDOM
    [ ~, err ] = cholcov(C_ab);
    if err == 0
        sample = mvnrnd(mu_ab, C_ab);
        sample = sample';
    else
        disp('WRONG COVARIANCE in predict_part_conditioned(RANDOM)-- using mean')
        sample = mu_ab;
    end
else
    sample = mu_ab;
end

z_j = sample(depCoeffInd);
if numel(sin_j) && numel(cos_j)
    qx = sample(QxDepInd - 2);
    qy = sample(QyDepInd - 2);
    sin_jOut = sin_j;
    cos_jOut = cos_j;
else
    qx = sample(QxDepInd);
    qy = sample(QyDepInd);
    sin_jOut = sample(sinDepInd);
    cos_jOut = sample(cosDepInd);
end

%% Compute the part center. I need the joint location of both parts in
%% their local coordinate system
B = model{modelInd}.eigVectors{1};
m = model{modelInd}.eigMean{1};

if isempty(z_i)
    points = m;
else
    points = B(:,1:nBref)*z_i+m;
end

ind = this.jointConnectionMatrix(model{modelInd}.depPartInd, model{modelInd}.refPartInd);
Qix_i = points(2*model{modelInd}.nPoints{1}+2*(ind-1)+1);
Qiy_i = points(2*model{modelInd}.nPoints{1}+2*(ind-1)+2);

B = model{modelInd}.eigVectors{2};
m = model{modelInd}.eigMean{2};

if isempty(z_j)
    points = m;
else
    points = B(:,1:nBdep)*z_j+m;
end
ind = this.jointConnectionMatrix(model{modelInd}.refPartInd, model{modelInd}.depPartInd);
Qjx_j = points(2*model{modelInd}.nPoints{2}+2*(ind-1)+1);
Qjy_j = points(2*model{modelInd}.nPoints{2}+2*(ind-1)+2);

Qjx_i = Qix_i + qx;
Qjy_i = Qiy_i + qy;

Ojx_j = -Qjx_j;
Ojy_j = -Qjy_j;

Angle = atan2(sin_jOut, cos_jOut);
[ Ojx_i Ojy_i ] = object_to_world(Ojx_j, Ojy_j, Angle, Qjx_i, Qjy_i);



end

